/*
  License text attribution: zlib license at

  http://www.gzip.org/zlib/zlib_license.html

  This license is a modification of the zlib license, changing names and
  dates.

  Copyright (C) 2007 Samson Chee

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Samson Chee samson@herethere.net
*/

// bubble game.cpp : Defines the entry point for the application.

#include "stdafx.h"

#include <stdio.h>
#include <time.h>
#include <ShellAPI.h>

#include "license.h"

#include "bubble game.h"
#include "bubble_data.h"
#include "bubble_manager.h"
#include "bubble_sprite.h"

#define MAX_LOADSTRING 100

#define kWinDimX 320
#define kWinDimY 320

#define kBoxSize 23
#define kBoxBorder 1

#ifndef NDEBUG
#define kTimerDelay 0
#else
#define kTimerDelay 40
#endif

#define kDropSpeed 8
#define kDropAcceleration 2.0f

// Global Variables:
HINSTANCE hInst;								// current instance
TCHAR szTitle[MAX_LOADSTRING];					// The title bar text
TCHAR szWindowClass[MAX_LOADSTRING];			// the main window class name

sBubbleManager* theBM;
int theTimerSequence = -1;
int theTimerSubsequence = 0;
bool theAutohint = false;
bool theAutorun = false;
bool theAnimated = true;
int theHintMode = 5;

HBITMAP theSprite[kBubbleDataMaxColors] = {0, 0, 0, 0, 0, 0, 0, 0};
HDC theSpriteDC[kBubbleDataMaxColors] = {0, 0, 0, 0, 0, 0, 0, 0};
HGDIOBJ theSpriteDCOld[kBubbleDataMaxColors] = {0, 0, 0, 0, 0, 0, 0 ,0};

// Forward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

void flipRect(RECT&);

void handleNew(HWND);
void handleUndo(HWND);
void handleRedo(HWND);
void handleHint(HWND);
void handleHintHelper(HWND);
void handleAutohint(HWND);
void handleAutorun(HWND);
void handleAnimated(HWND);
void handleQx(HWND);
void handleQy(HWND);
void handleFullGravityX(HWND);
void handleColors(HWND, int);
void handleHintMode(HWND, int);
void handleScoreMode(HWND, int);
void handlePaint(HWND);
void handleDestroy();
void handleClick(HWND, int, int);
void handleTimer(HWND, bool);

void beginAnimation(HWND);
void finishAnimation(HWND);

void drawBDClearMoving(HDC);
void drawBDClearBorders(HDC);
void drawBDBasic(HDC);
void drawBDSelect(HDC);

int getIndexByColor(tBDColor pColor);
COLORREF getRGBbyIndex(int pIndex);
COLORREF getRGBbyColor(tBDColor pColor);

int APIENTRY _tWinMain(HINSTANCE hInstance,
                       HINSTANCE hPrevInstance,
                       LPTSTR    lpCmdLine,
                       int       nCmdShow)
{
  // don't want to wait for animation in debug mode
#ifndef NDEBUG
  theAnimated = false;
#endif

  // init rng
#ifndef NDEBUG
  srand(0);
#else
  srand((unsigned int)time(0));
#endif
  theBM = bm_init();

  if (!theBM)
  {
    MessageBox(0, "Copyright (C) 2007 - Samson Chee", "Copyright", MB_OK);
  }

  // init sprites
  int theMapSize = kBoxSize - (2 * kBoxBorder);

  float* theBubbleMap = bs_calculate_map(theMapSize, theMapSize);
  for (int i = 0; i < kBubbleDataMaxColors; i++)
  {
    theSprite[i] = bs_color_sphere(0, theMapSize, theMapSize, theBubbleMap, getRGBbyIndex(i+1));
    theSpriteDC[i] = CreateCompatibleDC(0);
    theSpriteDCOld[i] = SelectObject(theSpriteDC[i], theSprite[i]);
  }
  bs_free_map(theBubbleMap);

  // TODO: Place code here.
  MSG msg;
  HACCEL hAccelTable;

  // Initialize global strings
  LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
  LoadString(hInstance, IDC_BUBBLEGAME, szWindowClass, MAX_LOADSTRING);
  MyRegisterClass(hInstance);

  // Perform application initialization:
  if (!InitInstance (hInstance, nCmdShow)) 
  {
    return FALSE;
  }

  hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_BUBBLEGAME);

  // Main message loop:
  while (GetMessage(&msg, NULL, 0, 0)) 
  {
    if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  return (int) msg.wParam;
}



//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage are only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
  WNDCLASSEX wcex;

  wcex.cbSize = sizeof(WNDCLASSEX); 

  wcex.style			= CS_HREDRAW | CS_VREDRAW;
  wcex.lpfnWndProc	= (WNDPROC)WndProc;
  wcex.cbClsExtra		= 0;
  wcex.cbWndExtra		= 0;
  wcex.hInstance		= hInstance;
  wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_BUBBLEGAME);
  wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
  //wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
  wcex.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
  wcex.lpszMenuName	= (LPCTSTR)IDC_BUBBLEGAME;
  wcex.lpszClassName	= szWindowClass;
  wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

  return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HANDLE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
  HWND hWnd;

  hInst = hInstance; // Store instance handle in our global variable

  hWnd = CreateWindow(
    szWindowClass,
    szTitle,
    WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
    CW_USEDEFAULT,
    CW_USEDEFAULT,
    kWinDimX,
    kWinDimY,
    NULL,
    NULL,
    hInstance,
    NULL);

  if (!hWnd)
  {
    return FALSE;
  }

  // before we call handleAnimated the first time, flip the animation state
  // because handleAnimated flips it by itself already
  theAnimated = !theAnimated;

  // flip game states the same way as above
  bm_gm_set_use_qx(theBM, !bm_gm_get_use_qx(theBM));
  bm_gm_set_use_qy(theBM, !bm_gm_get_use_qy(theBM));
  bm_gm_set_full_gravity_x(theBM, !bm_gm_get_full_gravity_x(theBM));

  // initialization handlers
  handleNew(hWnd);
  handleAnimated(hWnd);
  handleQx(hWnd);
  handleQy(hWnd);
  handleFullGravityX(hWnd);
  handleColors(hWnd, theBM->mGameMode.mColors);
  handleHintMode(hWnd, theHintMode);
  handleScoreMode(hWnd, bm_gm_get_score_mode(theBM));

  // prepare to show window
  RECT theRect;
  GetClientRect(hWnd, &theRect);
  SetWindowPos(
    hWnd,
    HWND_TOP,
    0, 0,
    kWinDimX + (kWinDimX - theRect.right), kWinDimY + (kWinDimY - theRect.bottom),
    SWP_NOMOVE);

  ShowWindow(hWnd, nCmdShow);
  UpdateWindow(hWnd);

  return TRUE;
}

//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  int wmId, wmEvent;

  switch (message) 
  {
  case WM_COMMAND:
    wmId    = LOWORD(wParam); 
    wmEvent = HIWORD(wParam); 
    // Parse the menu selections:
    switch (wmId)
    {
    case IDM_ABOUT:
      DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
      break;
    case ID_FILE_NEW:
      handleNew(hWnd);
      break;
    case ID_TOOLS_UNDO:
      handleUndo(hWnd);
      break;
    case ID_TOOLS_REDO:
      handleRedo(hWnd);
      break;
    case ID_TOOLS_HINT:
      handleHint(hWnd);
      break;
    case ID_TOOLS_AUTOHINT:
      handleAutohint(hWnd);
      break;
    case ID_TOOLS_AUTORUN:
      handleAutorun(hWnd);
      break;
    case ID_MODE_ANIMATED:
      handleAnimated(hWnd);
      break;
    case ID_MODE_QX:
      handleQx(hWnd);
      break;
    case ID_MODE_QY:
      handleQy(hWnd);
      break;
    case ID_MODE_FULLGRAVITYX:
      handleFullGravityX(hWnd);
      break;
    case ID_COLORS_2:
      handleColors(hWnd, 2);
      break;
    case ID_COLORS_3:
      handleColors(hWnd, 3);
      break;
    case ID_COLORS_4:
      handleColors(hWnd, 4);
      break;
    case ID_COLORS_5:
      handleColors(hWnd, 5);
      break;
    case ID_COLORS_6:
      handleColors(hWnd, 6);
      break;
    case ID_COLORS_7:
      handleColors(hWnd, 7);
      break;
    case ID_HINT_MODE1:
      handleHintMode(hWnd, 1);
      break;
    case ID_HINT_MODE2:
      handleHintMode(hWnd, 2);
      break;
    case ID_HINT_MODE3:
      handleHintMode(hWnd, 3);
      break;
    case ID_HINT_MODE4:
      handleHintMode(hWnd, 4);
      break;
    case ID_HINT_MODE5:
      handleHintMode(hWnd, 5);
      break;
    case ID_HINT_MODE6:
      handleHintMode(hWnd, 6);
      break;
    case ID_HINT_MODE7:
      handleHintMode(hWnd, 7);
      break;
    case ID_SCORING_MODE1:
      handleScoreMode(hWnd, 1);
      break;
    case ID_SCORING_MODE2:
      handleScoreMode(hWnd, 2);
      break;
    case ID_SCORING_MODE3:
      handleScoreMode(hWnd, 3);
      break;
    case ID_SCORING_MODE4:
      handleScoreMode(hWnd, 4);
      break;
    case ID_SCORING_MODE5:
      handleScoreMode(hWnd, 5);
      break;
    case ID_SCORING_MODE6:
      handleScoreMode(hWnd, 6);
      break;
    case IDM_EXIT:
      DestroyWindow(hWnd);
      break;
    default:
      return DefWindowProc(hWnd, message, wParam, lParam);
    }
    break;
  case WM_PAINT:
    handlePaint(hWnd);
    break;
  case WM_DESTROY:
    handleDestroy();
    break;
  case WM_LBUTTONUP:
    handleClick(hWnd, LOWORD(lParam), HIWORD(lParam));
    break;
  case WM_TIMER:
    handleTimer(hWnd, true);
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return 0;
}

// Message handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
  case WM_INITDIALOG:
    {
      int theVerLo;
      int theVerHi;

      UINT theDataSize;

      VS_FIXEDFILEINFO* theVersionInfo;
      char* theCompanyName;
      char* theProductName;
      char* theLegalCopyright;
      char* theSpecialBuild;

      char theAppName[2048];
      GetModuleFileName(NULL, theAppName, sizeof(theAppName));
      DWORD theFileVersionInfoSize  = GetFileVersionInfoSize(theAppName, 0);

      void* theVerBuf = new char[theFileVersionInfoSize];
      
      GetFileVersionInfo(theAppName, 0, theFileVersionInfoSize, theVerBuf);
      VerQueryValue(theVerBuf, "\\", (void**)&theVersionInfo, &theDataSize);
      VerQueryValue(theVerBuf, TEXT("\\StringFileInfo\\040904B0\\CompanyName"), (void**)&theCompanyName, &theDataSize);
      VerQueryValue(theVerBuf, TEXT("\\StringFileInfo\\040904B0\\ProductName"), (void**)&theProductName, &theDataSize);
      VerQueryValue(theVerBuf, TEXT("\\StringFileInfo\\040904B0\\LegalCopyright"), (void**)&theLegalCopyright, &theDataSize);
      VerQueryValue(theVerBuf, TEXT("\\StringFileInfo\\040904B0\\SpecialBuild"), (void**)&theSpecialBuild, &theDataSize);

      theVerLo = (theVersionInfo->dwFileVersionMS >> 0x00) & 0xffff;
      theVerHi = (theVersionInfo->dwFileVersionMS >> 0x10) & 0xffff;

      char theVersionStr[2048];
      sprintf(theVersionStr, "%s - version %d.%02d - %s", theProductName, theVerHi, theVerLo, theSpecialBuild);

      char theCopyrightStr[2048];
      sprintf(theCopyrightStr, "%s - %s", theLegalCopyright, theCompanyName);

      SetDlgItemText(hDlg, IDC_STATIC_VERSION, theVersionStr);
      SetDlgItemText(hDlg, IDC_STATIC_LEGALCOPYRIGHT, theCopyrightStr);
      SetDlgItemText(hDlg, IDC_EDIT_LICENSE, kLicenseText);

      delete theVerBuf;

      return TRUE;
    }
    break;

  case WM_COMMAND:
    {
      int wmId, wmEvent;

      wmId    = LOWORD(wParam); 
      wmEvent = HIWORD(wParam); 
      // Parse the menu selections:
      switch (wmId)
      {
      case IDC_BUTTON_GOWEB:
        ShellExecute(NULL, "open", "http://www.herethere.net/bg", NULL, NULL, SW_SHOWNORMAL);;
        break;
      }

  case IDM_ABOUT:
    if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
    {
      EndDialog(hDlg, LOWORD(wParam));
      return TRUE;
    }
    }
    break;
  }
  return FALSE;
}

void flipRect(RECT& pRect)
{
  pRect.bottom = kWinDimY - pRect.bottom;
  pRect.left = kWinDimX - pRect.left;
  pRect.right = kWinDimX - pRect.right;
  pRect.top = kWinDimY - pRect.top;
}

void handleNew(HWND pHWnd)
{
  finishAnimation(pHWnd);

  // re-randomize game data
  bm_new_game(theBM);

  // refresh display
  InvalidateRect(pHWnd, 0, FALSE);

  HMENU theMenu = GetMenu(pHWnd);
  EnableMenuItem(theMenu, ID_TOOLS_UNDO, MF_BYCOMMAND | (bm_can_undo(theBM) ? MF_ENABLED : MF_GRAYED));
  EnableMenuItem(theMenu, ID_TOOLS_REDO, MF_BYCOMMAND | (bm_can_redo(theBM) ? MF_ENABLED : MF_GRAYED));
}

void handleUndo(HWND pHWnd)
{
  if (bm_can_undo(theBM))
  {
    finishAnimation(pHWnd);

    // undo
    bm_undo(theBM);

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);

    HMENU theMenu = GetMenu(pHWnd);
    EnableMenuItem(theMenu, ID_TOOLS_UNDO, MF_BYCOMMAND | (bm_can_undo(theBM) ? MF_ENABLED : MF_GRAYED));
    EnableMenuItem(theMenu, ID_TOOLS_REDO, MF_BYCOMMAND | (bm_can_redo(theBM) ? MF_ENABLED : MF_GRAYED));
  }
}

void handleRedo(HWND pHWnd)
{
  if (bm_can_redo(theBM))
  {
    finishAnimation(pHWnd);

    // undo
    bm_redo(theBM);

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);

    HMENU theMenu = GetMenu(pHWnd);
    EnableMenuItem(theMenu, ID_TOOLS_UNDO, MF_BYCOMMAND | (bm_can_undo(theBM) ? MF_ENABLED : MF_GRAYED));
    EnableMenuItem(theMenu, ID_TOOLS_REDO, MF_BYCOMMAND | (bm_can_redo(theBM) ? MF_ENABLED : MF_GRAYED));
  }
}

void handleHint(HWND pHWnd)
{
  finishAnimation(pHWnd);
  handleHintHelper(pHWnd);
}

void handleHintHelper(HWND pHWnd)
{
  if (bm_current_getNumSelected(theBM) > 1)
  {
    bm_current_mark_clear(theBM);
    beginAnimation(pHWnd);
  }
  else
  {
    bm_current_mark_hint_group(theBM, theHintMode);

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void handleAutohint(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, ID_TOOLS_AUTOHINT, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  theAutohint = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_TOOLS_AUTOHINT, theNewMenuState);

  if (theAutohint && !theAutorun)
  {
    finishAnimation(pHWnd);

    if (bm_current_getNumSelected(theBM) <= 1)
    {
      bm_current_mark_hint_group(theBM, theHintMode);

      // refresh display
      InvalidateRect(pHWnd, 0, FALSE);
    }
  }
}

void handleAutorun(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, ID_TOOLS_AUTORUN, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  theAutorun = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_TOOLS_AUTORUN, theNewMenuState);

  if (theAutorun)
  {
    handleHintHelper(pHWnd);

    theTimerSequence = 6;
    SetTimer(pHWnd, 0, kTimerDelay, 0);
  }
}

void handleAnimated(HWND pHWnd)
{
  theAnimated = !theAnimated;

  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_MODE_ANIMATED, theAnimated ? MF_CHECKED : MF_UNCHECKED);
}

void handleQx(HWND pHWnd)
{
  bm_gm_set_use_qx(theBM, !bm_gm_get_use_qx(theBM));

  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_MODE_QX, bm_gm_get_use_qx(theBM) ? MF_CHECKED : MF_UNCHECKED);

  // refresh display
  InvalidateRect(pHWnd, 0, TRUE);
}

void handleQy(HWND pHWnd)
{
  bm_gm_set_use_qy(theBM, !bm_gm_get_use_qy(theBM));

  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_MODE_QY, bm_gm_get_use_qy(theBM) ? MF_CHECKED : MF_UNCHECKED);

  // refresh display
  InvalidateRect(pHWnd, 0, TRUE);
}

void handleFullGravityX(HWND pHWnd)
{
  bm_gm_set_full_gravity_x(theBM, !bm_gm_get_full_gravity_x(theBM));

  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_MODE_FULLGRAVITYX, bm_gm_get_full_gravity_x(theBM) ? MF_CHECKED : MF_UNCHECKED);
}

void handleColors(HWND pHWnd, int pMode)
{
  UINT theMenuId = 0;

  switch (pMode)
  {
  case 2: theMenuId = ID_COLORS_2; break;
  case 3: theMenuId = ID_COLORS_3; break;
  case 4: theMenuId = ID_COLORS_4; break;
  case 5: theMenuId = ID_COLORS_5; break;
  case 6: theMenuId = ID_COLORS_6; break;
  case 7: theMenuId = ID_COLORS_7; break;
  default: return;
  }

  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, theMenuId, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  bool theChecked = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_COLORS_2, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_COLORS_3, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_COLORS_4, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_COLORS_5, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_COLORS_6, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_COLORS_7, MF_UNCHECKED);

  CheckMenuItem(theMenu, theMenuId, theNewMenuState);

  theBM->mGameMode.mColors = pMode;
}

void handleHintMode(HWND pHWnd, int pMode)
{
  UINT theMenuId = 0;

  switch (pMode)
  {
  case 1: theMenuId = ID_HINT_MODE1; break;
  case 2: theMenuId = ID_HINT_MODE2; break;
  case 3: theMenuId = ID_HINT_MODE3; break;
  case 4: theMenuId = ID_HINT_MODE4; break;
  case 5: theMenuId = ID_HINT_MODE5; break;
  case 6: theMenuId = ID_HINT_MODE6; break;
  case 7: theMenuId = ID_HINT_MODE7; break;
  }

  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, theMenuId, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  bool theChecked = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_HINT_MODE1, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE2, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE3, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE4, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE5, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE6, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_HINT_MODE7, MF_UNCHECKED);

  CheckMenuItem(theMenu, theMenuId, theNewMenuState);

  theHintMode = pMode;
}

void handleScoreMode(HWND pHWnd, int pMode)
{
  UINT theMenuId = 0;

  switch (pMode)
  {
  case 1: theMenuId = ID_SCORING_MODE1; break;
  case 2: theMenuId = ID_SCORING_MODE2; break;
  case 3: theMenuId = ID_SCORING_MODE3; break;
  case 4: theMenuId = ID_SCORING_MODE4; break;
  case 5: theMenuId = ID_SCORING_MODE5; break;
  case 6: theMenuId = ID_SCORING_MODE6; break;
  }

  HMENU theMenu = GetMenu(pHWnd);

  UINT theMenuState = GetMenuState(theMenu, theMenuId, MF_BYCOMMAND);
  UINT theNewMenuState = (theMenuState & MF_CHECKED) ? MF_UNCHECKED : MF_CHECKED;
  bool theChecked = (theNewMenuState == MF_CHECKED);

  CheckMenuItem(theMenu, ID_SCORING_MODE1, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_SCORING_MODE2, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_SCORING_MODE3, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_SCORING_MODE4, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_SCORING_MODE5, MF_UNCHECKED);
  CheckMenuItem(theMenu, ID_SCORING_MODE6, MF_UNCHECKED);

  CheckMenuItem(theMenu, theMenuId, theNewMenuState);

  bm_gm_set_score_mode(theBM, pMode);
}

void handlePaint(HWND pHWnd)
{
  PAINTSTRUCT thePS;
  HDC theHDC;

  // drawing initialize
  theHDC = BeginPaint(pHWnd, &thePS);
  int theOldRop = GetROP2(theHDC);

  // clear hdc
  drawBDClearMoving(theHDC);

  if (bm_current_getNumSelected(theBM) || (theTimerSequence < 0))
  {
    drawBDClearBorders(theHDC);
  }

  // drawing code
  HPEN thePen = CreatePen(PS_SOLID, kBoxBorder, RGB(0x00,0xff,0xff));
  HGDIOBJ theOldPen = SelectObject(theHDC, thePen);

  MoveToEx(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY), NULL);
  LineTo(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY);
  MoveToEx(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY), NULL);
  LineTo(theHDC, kWinDimX, kWinDimY - 1 - (kBoxSize * kBubbleDataMaxY));

  SelectObject(theHDC, theOldPen);
  DeleteObject(thePen);

  SetTextColor(theHDC, RGB(0xff, 0xff, 0xff));
  SetBkColor(theHDC, RGB(0x00, 0x00, 0x00));

  // remember to pad the ends of the output strings with spaces because it's easier to overwrite existing
  // text that way
  sBubbleData* theBD = bm_current_data(theBM);
  char theBuf[8192];

  {
    HFONT theFont = CreateFont(17, 0, 0, 0, 0, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, 0);
    HGDIOBJ theOldFont = SelectObject(theHDC, theFont);

    int theGroupSize = (theTimerSequence < 0) ? theBD->mNumSelected : theBD->mLastGroupSize;
    int theGroupScore = (theTimerSequence < 0) ? groupScore(&theBM->mGameMode, theBD->mNumSelected) : theBD->mLastGroupScore;

    sprintf(theBuf, "score: %d   current: %d/%d                ", theBD->mScore, theGroupSize, theGroupScore);
    TextOut(theHDC, kWinDimX - 1 - (kBoxSize * kBubbleDataMaxX), 4, theBuf, (int)strlen(theBuf));

    SelectObject(theHDC, theOldFont);
    DeleteObject(theFont);
  }

  {
    HFONT theFont = CreateFont(17, 0, 900, 0, 0, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, 0);
    HGDIOBJ theOldFont = SelectObject(theHDC, theFont);

    float theAverageScore = (theBD->mMoves < 1) ? 0 : (float)((theBD->mScore * 1.0) / (theBD->mMoves * 1.0));

    sprintf(theBuf, "moves: %d   average: %2.2f   largest: %d                ", theBD->mMoves, theAverageScore, theBD->mLargestGroup);
    TextOut(theHDC, 2, kWinDimY - 4, theBuf, (int)strlen(theBuf));

    SelectObject(theHDC, theOldFont);
    DeleteObject(theFont);
  }

  drawBDBasic(theHDC);
  drawBDSelect(theHDC);

  // drawing cleanup
  SetROP2(theHDC, theOldRop);
  EndPaint(pHWnd, &thePS);
}

void handleDestroy()
{
  bm_destroy(theBM);

  for (int i = 0; i < kBubbleDataMaxColors; i++)
  {
    if (theSprite[i])
    {
      SelectObject(theSpriteDC[i], theSpriteDCOld[i]);
      DeleteObject(theSprite[i]);
      DeleteDC(theSpriteDC[i]);
    }
  }

  PostQuitMessage(0);
}

void handleClick(HWND pHWnd, int pX, int pY)
{
  finishAnimation(pHWnd);

  int theX = kWinDimX - pX;
  int theY = kWinDimY - pY;

  int theGridX = theX / kBoxSize;
  int theGridY = theY / kBoxSize;

  if ((theGridX >= 0) && (theGridX < kBubbleDataMaxX) && (theGridY >= 0) && (theGridY < kBubbleDataMaxY))
  {
    if ((bm_current_getNumSelected(theBM) > 1) && bm_current_ismarked(theBM, theGridX, theGridY))
    {
      bm_current_mark_clear(theBM);
      beginAnimation(pHWnd);
    }
    else
    {
      bm_current_mark_select(theBM, theGridX, theGridY);
    }
  }
  else
  {
    bm_current_reset_mark(theBM);
  }

  // refresh display
  InvalidateRect(pHWnd, 0, FALSE);
}

void handleTimer(HWND pHWnd, bool pRefresh)
{
  sBubbleData* theBD = bm_current_data(theBM);

  do
    {
    switch (theTimerSequence)
    {
    case 0 :
      {
        bm_current_gravityY_pre(theBM);

        theTimerSubsequence = kDropSpeed;
        theTimerSequence++;
      }; break;
    case 1 :
      {
        theTimerSubsequence += (int)(theTimerSubsequence / kDropAcceleration);

        if (theTimerSubsequence >= theBD->mMaxMove * kBoxSize)
        {
          theTimerSequence++;
        }
      }; break;
    case 2 :
      {
        bm_current_gravityY_post(theBM);

        if (bm_current_can_load_qx(theBM))
        {
          bm_current_load_qx(theBM);
          bm_current_gravityY_pre(theBM);

          theTimerSubsequence = kDropSpeed;
          theTimerSequence = 1;
        }
        else
        {
          bm_current_gravityX_pre(theBM);

          theTimerSubsequence = kDropSpeed;
          theTimerSequence++;
        }
      }; break;
    case 3 :
      {
        theTimerSubsequence += (int)(theTimerSubsequence / kDropAcceleration);

        if (theTimerSubsequence >= theBD->mMaxMove * kBoxSize)
        {
          theTimerSequence++;
        }
      }; break;
    case 4 :
      {
        bm_current_gravityX_post(theBM);

        if (bm_current_can_load_qy(theBM))
        {
          bm_current_load_qy(theBM);
          bm_current_gravityX_pre(theBM);

          theTimerSubsequence = kDropSpeed;
          theTimerSequence = 3;
        }
        else
        {
          theTimerSubsequence = 0;
          theTimerSequence++;
        }
      }; break;
    case 5 :
      {
        KillTimer(pHWnd, 0);
        theTimerSequence = -1;

        // check game over
        if (!bm_current_game_active(theBM))
        {
          // refresh display
          InvalidateRect(pHWnd, 0, FALSE);

          // new game prompt
          if (theAutorun) {
  #ifndef NDEBUG
            char theBuf[8192];
            sprintf(theBuf, "count = %d; score = %d\n", theBD->mMoves, theBD->mScore);
            OutputDebugString(theBuf);
  #endif
            bm_new_game(theBM);

            HMENU theMenu = GetMenu(pHWnd);
            EnableMenuItem(theMenu, ID_TOOLS_UNDO, MF_BYCOMMAND | (bm_can_undo(theBM) ? MF_ENABLED : MF_GRAYED));
            EnableMenuItem(theMenu, ID_TOOLS_REDO, MF_BYCOMMAND | (bm_can_redo(theBM) ? MF_ENABLED : MF_GRAYED));
          } 
          else if (MessageBox(pHWnd, "Game Over - New Game?", "Game Over", MB_YESNO) == IDYES)
          {
            handleNew(pHWnd);
          }
        }

        if (theAutorun)
        {
          handleHintHelper(pHWnd);
          theTimerSequence = 6;

          // - if we're not animated, we'll go to case 6 in the next loop anyway
          // - if pRefresh is false, someone else is pumping the loop externally
          if (theAnimated && pRefresh)
          {
            SetTimer(pHWnd, 0, kTimerDelay, 0);
          }
        }
        else if (theAutohint)
        {
          handleHintHelper(pHWnd);
        }

      }; break;
    case 6 :
      {
        bm_current_mark_clear(theBM);
        beginAnimation(pHWnd);
      }; break;
    default :
      {
        // shouldn't get here, but if we do, put the timer sequence in a valid state
        theTimerSequence = -1;
      }
    }
  } while (!theAnimated && (theTimerSequence > 0));

  if (pRefresh)
  {
    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void beginAnimation(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);
  EnableMenuItem(theMenu, ID_TOOLS_UNDO, MF_BYCOMMAND | (bm_can_undo(theBM) ? MF_ENABLED : MF_GRAYED));
  EnableMenuItem(theMenu, ID_TOOLS_REDO, MF_BYCOMMAND | (bm_can_redo(theBM) ? MF_ENABLED : MF_GRAYED));

  theTimerSequence = 0;
  SetTimer(pHWnd, 0, kTimerDelay, 0);
}

void finishAnimation(HWND pHWnd)
{
  HMENU theMenu = GetMenu(pHWnd);
  CheckMenuItem(theMenu, ID_TOOLS_AUTORUN, MF_UNCHECKED);

  theAutorun = false;

  if (theTimerSequence >= 0)
  {
    KillTimer(pHWnd, 0);

    while (theTimerSequence >= 0)
    {
      handleTimer(pHWnd, false);
    }

    // refresh display
    InvalidateRect(pHWnd, 0, FALSE);
  }
}

void drawBDClearMoving(HDC pHDC)
{
  RECT theRect;
  HBRUSH theBackgroundBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  // matrix
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bm_current_ptr(theBM, i, j);
      if (theNode->mMark || theNode->mMove || (theNode->mColor == 0))
      {
        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        flipRect(theRect);

        FillRect(pHDC, &theRect, theBackgroundBrush);
      }
    }
  }

}

void drawBDClearBorders(HDC pHDC)
{
  RECT theRect;
  HBRUSH theBackgroundBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  for (tBDDim i = 0; i <= kBubbleDataMaxX; i++)
  {
    theRect.top = 0;
    theRect.bottom = kBubbleDataMaxY * kBoxSize;

    theRect.left = i * kBoxSize;
    theRect.left = (i == 0) ? theRect.left : theRect.left - kBoxBorder;

    theRect.right = theRect.left + kBoxBorder;
    theRect.right = (i == kBubbleDataMaxX) ? theRect.right : theRect.right + kBoxBorder;

    flipRect(theRect);
    FillRect(pHDC, &theRect, theBackgroundBrush);
  }

  for (tBDDim j = 0; j <= kBubbleDataMaxY; j++)
  {
    theRect.top = j * kBoxSize;
    theRect.top = (j == 0) ? theRect.top : theRect.top - kBoxBorder;

    theRect.bottom = theRect.top + kBoxBorder;
    theRect.bottom = (j == kBubbleDataMaxY) ? theRect.bottom : theRect.bottom + kBoxBorder;

    theRect.left = 0;
    theRect.right = kBubbleDataMaxX * kBoxSize;

    flipRect(theRect);
    FillRect(pHDC, &theRect, theBackgroundBrush);
  }

}

void drawBDBasic(HDC pHDC)
{
  RECT theRect;

  // matrix
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bm_current_ptr(theBM, i, j);
      if (theNode->mColor)
      {
        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        if (theNode->mMove)
        {
          switch (theTimerSequence)
          {
          case 1:
          case 2:
            {
              if (theTimerSubsequence < theNode->mMove * kBoxSize)
              {
                theRect.top -= theTimerSubsequence;
                theRect.bottom -= theTimerSubsequence;
              }
              else
              {
                theRect.top -= theNode->mMove * kBoxSize;
                theRect.bottom -= theNode->mMove * kBoxSize;
              }
            }; break;
          case 3:
          case 4:
            {
              if (theTimerSubsequence < theNode->mMove * kBoxSize)
              {
                theRect.left -= theTimerSubsequence;
                theRect.right -= theTimerSubsequence;
              }
              else
              {
                theRect.left-= theNode->mMove * kBoxSize;
                theRect.right -= theNode->mMove * kBoxSize;
              }
            }; break;
          }
        }

        InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
        flipRect(theRect);
        BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
      }
    }
  }

  // draw queues

  // x queue 
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bm_current_ptr_qx(theBM, i);

    theRect.top = kBubbleDataMaxY * kBoxSize;
    theRect.bottom = (kBubbleDataMaxY + 1) * kBoxSize;
    theRect.left = i * kBoxSize;
    theRect.right = (i + 1) * kBoxSize;

    InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
    flipRect(theRect);
    BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
  }

  // y queue
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    sBubbleNode* theNode = bm_current_ptr_qy(theBM, j);

    theRect.top = j * kBoxSize;
    theRect.bottom = (j + 1) * kBoxSize;
    theRect.left = kBubbleDataMaxX * kBoxSize;
    theRect.right = (kBubbleDataMaxX + 1) * kBoxSize;

    InflateRect(&theRect, -kBoxBorder, -kBoxBorder);
    flipRect(theRect);
    BitBlt(pHDC, theRect.right, theRect.bottom, theRect.left - theRect.right, theRect.top - theRect.bottom, theSpriteDC[getIndexByColor(theNode->mColor) - 1], 0, 0, SRCCOPY);
  }

}

void drawBDSelect(HDC pHDC)
{
  if (bm_current_getNumSelected(theBM) <= 1)
  {
    return;
  }

  HBRUSH thePosBrush = (HBRUSH)GetStockObject(WHITE_BRUSH);
  HBRUSH theNegBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);

  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bm_current_ptr(theBM, i, j);
      if (theNode->mMark)
      {
        RECT theRect;

        theRect.top = j * kBoxSize;
        theRect.bottom = (j + 1) * kBoxSize;
        theRect.left = i * kBoxSize;
        theRect.right = (i + 1) * kBoxSize;

        flipRect(theRect);

        //FillRect(pHDC, &theRect, thePosBrush);

        bool theDrawFollowing = false;

        // side cases
        theDrawFollowing = true;
        if (i > 0)
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i - 1, j)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (i < kBubbleDataMaxX - 1)
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i + 1, j)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (j > 0)
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if (j < kBubbleDataMaxY - 1)
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = (i + 1) * kBoxSize;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        // corner cases

        theDrawFollowing = true;
        if ((i > 0) && (j > 0))
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i - 1, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i > 0) && (j < kBubbleDataMaxY - 1))
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i - 1, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.left = i * kBoxSize;
          theRect.right = theRect.left + kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i < kBubbleDataMaxX - 1) && (j > 0))
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i + 1, j - 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.top = j * kBoxSize;
          theRect.bottom = theRect.top + kBoxBorder;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

        theDrawFollowing = true;
        if ((i < kBubbleDataMaxX - 1) && (j < kBubbleDataMaxY - 1))
        {
          theDrawFollowing = !(bm_current_ptr(theBM, i + 1, j + 1)->mMark);
        }

        if (theDrawFollowing)
        {
          theRect.bottom = (j + 1) * kBoxSize;
          theRect.top = theRect.bottom - kBoxBorder;
          theRect.right = (i + 1) * kBoxSize;
          theRect.left = theRect.right - kBoxBorder;

          flipRect(theRect);
          FillRect(pHDC, &theRect, thePosBrush);
        }

      }
    }
  }
}

int getIndexByColor(tBDColor pColor)
{
  switch (pColor)
  {
  case 0x00 : return 255;
  case 0xff : return 254;
  case 0x01 : return 1;
  case 0x02 : return 2;
  case 0x04 : return 3;
  case 0x08 : return 4;
  case 0x10 : return 5;
  case 0x20 : return 6;
  case 0x40 : return 7;
  case 0x80 : return 8;
  default : return 0;
  }
}

COLORREF getRGBbyIndex(int pIndex)
{
  switch (pIndex)
  {
  case 255 : return RGB(0xff,0xff,0xff);
  case 254 : return RGB(0x7f,0x7f,0x7f);
  case   1 : return RGB(0xff,0x00,0x00);
  case   2 : return RGB(0x00,0xff,0x00);
  case   3 : return RGB(0x7f,0x7f,0xff);
  case   4 : return RGB(0xff,0xd7,0x00);
  case   5 : return RGB(0xff,0x00,0xd7);
  case   6 : return RGB(0x00,0xff,0xff);
  case   7 : return RGB(0xaf,0xff,0xaf);
  case   8 : return RGB(0xff,0xaf,0xaf);
  default : return RGB(0x00,0x00,0x00);
  }
}

COLORREF getRGBbyColor(tBDColor pColor)
{
  return getRGBbyIndex(getIndexByColor(pColor));
}
