/*
  License text attribution: zlib license at

  http://www.gzip.org/zlib/zlib_license.html

  This license is a modification of the zlib license, changing names and
  dates.

  Copyright (C) 2007 Samson Chee

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Samson Chee samson@herethere.net
*/

#include "stdafx.h"

#include <stdio.h>
#include "bubble_data.h"
#include "bubble_manager.h"

sBubbleManager* bm_init()
{
  sBubbleManager* theData = (sBubbleManager*)malloc(sizeof(sBubbleManager) + (sizeof(sBubbleData*) * kBubbleManagerTotalBuffer));

  theData->mFirst = 0;
  theData->mCurrent = 0;
  theData->mLast = 0;

  theData->mGameMode.mColors = 5;
  theData->mGameMode.mUseQx = true;
  theData->mGameMode.mUseQy = true;
  theData->mGameMode.mFullGravityX = true;
  theData->mGameMode.mScoreMode = 2;

  // since we allocated room for the mData array in the preceding malloc, we don't need to do it again
  // - point mData to the first location beyond itself
  theData->mData = (sBubbleData**)(&theData->mData);
  theData->mData++;
  for (int i = 0; i < kBubbleManagerTotalBuffer; i++)
  {
    theData->mData[i] = bd_init();
  }

  return theData;
}

void bm_destroy(sBubbleManager* pBM)
{
  for (int i = 0; i < kBubbleManagerTotalBuffer; i++)
  {
    bd_destroy(pBM->mData[i]);
  }

  // free the malloc for the entire data structure
  // - don't need to free pBM->mData specifically because it was never specifically allocated
  free(pBM);
}

void bm_new_game(sBubbleManager* pBM)
{
  pBM->mFirst = 0;
  pBM->mCurrent = 0;
  pBM->mLast = 0;

  bd_new_game(&pBM->mGameMode, bm_current_data(pBM));
}

sBubbleData* bm_current_data(sBubbleManager* pBM)
{
  return pBM->mData[pBM->mCurrent % kBubbleManagerUndoBuffer];
}

sBubbleData* bm_temp_data(sBubbleManager* pBM, unsigned int pSlot)
{
  if (pSlot >= kBubbleManagerTempBuffer)
  {
    return 0;
  }

  return pBM->mData[kBubbleManagerUndoBuffer + pSlot];
}

sBubbleData* bm_undo(sBubbleManager* pBM)
{
  if (!bm_can_undo(pBM))
  {
    return 0;
  }

  pBM->mCurrent--;
  return bm_current_data(pBM);
}

sBubbleData* bm_redo(sBubbleManager* pBM)
{
  if (!bm_can_redo(pBM))
  {
    return 0;
  }

  pBM->mCurrent++;
  return bm_current_data(pBM);
}

bool bm_can_undo(sBubbleManager* pBM)
{
  return (pBM->mCurrent > pBM->mFirst);
}

bool bm_can_redo(sBubbleManager* pBM)
{
  return (pBM->mCurrent < pBM->mLast);
}

int bm_gm_get_colors(sBubbleManager* pBM)
{
  return pBM->mGameMode.mColors;
}

bool bm_gm_get_use_qx(sBubbleManager* pBM)
{
  return (pBM->mGameMode.mUseQx != 0);
}

bool bm_gm_get_use_qy(sBubbleManager* pBM)
{
  return (pBM->mGameMode.mUseQy != 0);
}

bool bm_gm_get_full_gravity_x(sBubbleManager* pBM)
{
  return (pBM->mGameMode.mFullGravityX != 0);
}

int bm_gm_get_score_mode(sBubbleManager* pBM)
{
  return pBM->mGameMode.mScoreMode;
}

void bm_gm_set_colors(sBubbleManager* pBM, int pNum)
{
  if ((pNum < 2) || (pNum > 7))
  {
    return;
  }

  pBM->mGameMode.mColors = pNum;
}

void bm_gm_set_use_qx(sBubbleManager* pBM, bool pState)
{
  pBM->mGameMode.mUseQx = (pState != 0);
  bd_randomize_color_qx(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_gm_set_use_qy(sBubbleManager* pBM, bool pState)
{
  pBM->mGameMode.mUseQy = (pState != 0);
  bd_randomize_color_qy(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_gm_set_full_gravity_x(sBubbleManager* pBM, bool pState)
{
  pBM->mGameMode.mFullGravityX = (pState != 0);
}

void bm_gm_set_score_mode(sBubbleManager* pBM, int pMode)
{
  if ((pMode < 1) || (pMode > 6))
  {
    return;
  }

  pBM->mGameMode.mScoreMode = pMode;
}

tBDDim2 bm_current_getNumSelected(sBubbleManager* pBM)
{
  return bm_current_data(pBM)->mNumSelected;
}

tBDBool bm_current_game_active(sBubbleManager* pBM)
{
  return bd_game_active(bm_current_data(pBM));
}

sBubbleNode* bm_current_ptr(sBubbleManager* pBM, tBDDim pX, tBDDim pY)
{
  return bd_ptr(bm_current_data(pBM), pX, pY);
}

void bm_current_reset_mark(sBubbleManager* pBM)
{
  bd_reset_mark(bm_current_data(pBM));
}

tBDBool bm_current_ismarked(sBubbleManager* pBM, tBDDim pX, tBDDim pY)
{
  return bd_ismarked(bm_current_data(pBM), pX, pY);
}

tBDDim2 bm_current_mark_select(sBubbleManager* pBM, tBDDim pX, tBDDim pY)
{
  return bd_mark_select(bm_current_data(pBM), pX, pY);
}

tBDDim2 bm_current_mark_clear(sBubbleManager* pBM)
{
  bd_copy(pBM->mData[(pBM->mCurrent + 1) % kBubbleManagerUndoBuffer], pBM->mData[pBM->mCurrent % kBubbleManagerUndoBuffer]);

  pBM->mCurrent++;

  if (pBM->mCurrent - pBM->mFirst >= kBubbleManagerUndoBuffer)
  {
    pBM->mFirst = (pBM->mCurrent - kBubbleManagerUndoBuffer) + 1;
  }

  pBM->mLast = pBM->mCurrent;

  return bd_mark_clear(&pBM->mGameMode, bm_current_data(pBM));
}

tBDDim2 bm_current_mark_hint_group(sBubbleManager* pBM, int pMode)
{
  return bd_mark_hint_group(&pBM->mGameMode, bm_current_data(pBM), pMode, bm_temp_data(pBM));
}

void bm_current_gravityX_pre(sBubbleManager* pBM)
{
  bd_gravityX_pre(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_current_gravityX_post(sBubbleManager* pBM)
{
  bd_gravityX_post(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_current_gravityY_pre(sBubbleManager* pBM)
{
  bd_gravityY_pre(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_current_gravityY_post(sBubbleManager* pBM)
{
  bd_gravityY_post(&pBM->mGameMode, bm_current_data(pBM));
}

sBubbleNode* bm_current_ptr_qx(sBubbleManager* pBM, tBDDim pNum)
{
  return bd_ptr_qx(bm_current_data(pBM), pNum);
}

sBubbleNode* bm_current_ptr_qy(sBubbleManager* pBM, tBDDim pNum)
{
  return bd_ptr_qy(bm_current_data(pBM), pNum);
}

tBDBool bm_current_can_load_qx(sBubbleManager* pBM)
{
  return bd_can_load_qx(&pBM->mGameMode, bm_current_data(pBM));
}

tBDBool bm_current_can_load_qy(sBubbleManager* pBM)
{
  return bd_can_load_qy(&pBM->mGameMode, bm_current_data(pBM));
}

void bm_current_load_qx(sBubbleManager* pBM)
{
  bd_load_qx(&pBM->mGameMode, bm_current_data(pBM), 1);
}

void bm_current_load_qy(sBubbleManager* pBM)
{
  bd_load_qy(&pBM->mGameMode, bm_current_data(pBM), 1);
}
