/*
  License text attribution: zlib license at

  http://www.gzip.org/zlib/zlib_license.html

  This license is a modification of the zlib license, changing names and
  dates.

  Copyright (C) 2007 Samson Chee

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Samson Chee samson@herethere.net
*/

#ifndef __bubble_manager_h__
#define __bubble_manager_h__

#define kBubbleManagerUndoBuffer 256
#define kBubbleManagerTempBuffer 4
#define kBubbleManagerTotalBuffer (kBubbleManagerUndoBuffer + kBubbleManagerTempBuffer)

// container
struct sBubbleManager
{
  int mFirst;
  int mCurrent;
  int mLast;

  sBubbleMode mGameMode;

  sBubbleData** mData;
};

// functions
sBubbleManager* bm_init();
void bm_destroy(sBubbleManager* pBM);

void bm_new_game(sBubbleManager* pBM);
sBubbleData* bm_current_data(sBubbleManager* pBM);
sBubbleData* bm_temp_data(sBubbleManager* pBM, unsigned int pSlot = 0);

sBubbleData* bm_undo(sBubbleManager* pBM);
sBubbleData* bm_redo(sBubbleManager* pBM);

bool bm_can_undo(sBubbleManager* pBM);
bool bm_can_redo(sBubbleManager* pBM);

// -----
// game modes
// -----

int bm_gm_get_colors(sBubbleManager* pBM);
bool bm_gm_get_use_qx(sBubbleManager* pBM);
bool bm_gm_get_use_qy(sBubbleManager* pBM);
bool bm_gm_get_full_gravity_x(sBubbleManager* pBM);
int bm_gm_get_score_mode(sBubbleManager* pBM);

void bm_gm_set_colors(sBubbleManager* pBM, int pNum);
void bm_gm_set_use_qx(sBubbleManager* pBM, bool pState);
void bm_gm_set_use_qy(sBubbleManager* pBM, bool pState);
void bm_gm_set_full_gravity_x(sBubbleManager* pBM, bool pState);
void bm_gm_set_score_mode(sBubbleManager* pBM, int pMode);

// -----
// property get of current bubble data
// -----

tBDDim2 bm_current_getNumSelected(sBubbleManager* pBM);

// -----
// function invocation of current bubble data
// -----

tBDBool bm_current_game_active(sBubbleManager* pBM);

sBubbleNode* bm_current_ptr(sBubbleManager* pBM, tBDDim pX, tBDDim pY);

void bm_current_reset_mark(sBubbleManager* pBM);
tBDBool bm_current_ismarked(sBubbleManager* pBM, tBDDim pX, tBDDim pY);
tBDDim2 bm_current_mark_select(sBubbleManager* pBM, tBDDim pX, tBDDim pY);
tBDDim2 bm_current_mark_clear(sBubbleManager* pBM);

tBDDim2 bm_current_mark_hint_group(sBubbleManager* pBM, int pMode);

void bm_current_gravityX_pre(sBubbleManager* pBM);
void bm_current_gravityX_post(sBubbleManager* pBM);
void bm_current_gravityY_pre(sBubbleManager* pBM);
void bm_current_gravityY_post(sBubbleManager* pBM);

sBubbleNode* bm_current_ptr_qx(sBubbleManager* pBM, tBDDim pNum);
sBubbleNode* bm_current_ptr_qy(sBubbleManager* pBM, tBDDim pNum);

tBDBool bm_current_can_load_qx(sBubbleManager* pBM);
tBDBool bm_current_can_load_qy(sBubbleManager* pBM);

void bm_current_load_qx(sBubbleManager* pBM);
void bm_current_load_qy(sBubbleManager* pBM);

#endif // __bubble_manager_h__
